#include-once

#cs ================================================================================================
This UDF incapsulate all work with content plugins
#ce ================================================================================================

#cs ================================================================================================
2024-03-20 - tiny tweaks and cleanups
2024-02-18 - support for x64
2024-02-16 - update for support ft_multiplechoice and ft_fulltext 
#ce ================================================================================================

Global Const $MAX_COLUMNS = 256
Global Const $MAX_BUFFER_SIZE = 65534
Global Enum $PLUGIN_HANDLE, $PLUGIN_COLUMNS, $WDX_PLUGIN_NAME, $WDX_PLUGIN_PARAMS
Global Enum $COLUMN_TYPE, $COLUMN_NAME, $COLUMN_UNIT, $COLUMN_MAXINDEX
Global Enum $ERROR_ERROPEN = 1, $ERROR_ERRHANDLE, $ERROR_ERRZEROCOLUMNS, $ERROR_ERRCOLUMN, $ERROR_ERRUNIT, $ERROR_ERRCALL, $ERROR_ERRVALUE

; $ERROR_ERROPEN - error openinh of DLL
; $ERROR_ERRHANDLE - errorous "handle"
; $ERROR_ERRZEROCOLUMNS - no columns was retrieved
; $ERROR_ERRCOLUMN - error in column name/index
; $ERROR_ERRUNIT - error in unit name/index
; $ERROR_ERRCALL - error calling function (usually absent function)
; $ERROR_ERRVALUE - error retrieving value (wrong or unsupported type), @extended contain original return value

Global Const $ft_nomorefields = 0 
Global Const $ft_numeric_32 = 1 
Global Const $ft_numeric_64 = 2 
Global Const $ft_numeric_floating = 3 
Global Const $ft_date = 4 
Global Const $ft_time = 5 
Global Const $ft_boolean = 6 
Global Const $ft_multiplechoice = 7 
; $ft_string conflicts with WinAPI constant
Global Const $ft_stringA = 8
Global Const $ft_fulltext = 9 
Global Const $ft_datetime = 10 
Global Const $ft_stringW = 11 
Global Const $ft_fulltextW = 12 

Global Const $ft_comparecontent = 100

Global Const $ft_nosuchfield = -1
Global Const $ft_fileerror = -2
Global Const $ft_fieldempty = -3
Global Const $ft_ondemand = -4

Global Const $tagContentDefaultParamStruct = "int;dword;dword;char[256]"



; Result: special "handle" (acually array: [plugin handle, fields array])
; Error: 0 and sets @error
Func _WDXPluginOpen($sPlugin)
    If StringRight($sPlugin, 2) = 64 Then $sPlugin = StringTrimRight($sPlugin, 2)
    
    Local $bHas32 = FileExists($sPlugin), $bHas64 = FileExists($sPlugin & "64")
    
    Local $aParams[4] = [$sPlugin, $bHas32, $bHas64]
    
    If @AutoItX64 Then 
        If Not $bHas64 Then Return SetError($ERROR_ERROPEN, 1, 0)
        $sPlugin &= "64"
    Else
        If Not $bHas32 Then Return SetError($ERROR_ERROPEN, 1, 0)
    EndIf

    Local $hPlugin = DllOpen($sPlugin)
    If @Error Then Return SetError($ERROR_ERROPEN, 2, 0)

    Local $pDPS = DllStructCreate($tagContentDefaultParamStruct)
    ; Struct size
    DllStructSetData($pDPS, 1, DllStructGetSize($pDPS))
    ; Plugin version 2.11
    DllStructSetData($pDPS, 2, 11)
    DllStructSetData($pDPS, 3, 2)
    ; "Recommended" settings file path
    DllStructSetData($pDPS, 4, EnvGet("COMMANDER_PATH") & "\contplug.ini")
    DllCall($hPlugin, "none", "ContentSetDefaultParams", "ptr", DllStructGetPtr($pDPS))

    ; Get columns
    Local $aColumns = __WDXGetColumnsInternal($hPlugin)
    If @error Then Return SetError(@error, 0, 0)

    Local $sPlugName = StringRegExpReplace($sPlugin, "(?m)^.+?\\([^\\]+)\.[^\.]+$", "\1")
    Local $aPlugin[] = [$hPlugin, $aColumns, $sPlugName, $aParams]
    Return $aPlugin
EndFunc

; Result: none
Func _WDXPluginClose(ByRef $aPlugin)
    If Not _WDXIsValidHandle($aPlugin) Then Return

    DllCall($aPlugin[$PLUGIN_HANDLE], "none", "ContentPluginUnloading")
    DllClose($aPlugin[$PLUGIN_HANDLE])
    $aPlugin = 0
EndFunc

; Result: 2D array of all columns: [type, name, units], units are sub-array
; Error: empty string and sets @error
Func _WDXGetColumns(ByRef $aPlugin)
    If Not _WDXIsValidHandle($aPlugin) Then Return SetError($ERROR_ERRHANDLE, 0, "")

    Return $aPlugin[$PLUGIN_COLUMNS]
EndFunc

; Return column type, units or name
; Error: empty string and sets @error
Func _WDXGetColumnInfo(ByRef $aPlugin, $vColumn, $nInfoIndex)
    If Not _WDXIsValidHandle($aPlugin) Then Return SetError($ERROR_ERRHANDLE, 0, "")
    
    Local $aColumns = $aPlugin[$PLUGIN_COLUMNS]

    ; if field is given by name
    If Not StringIsDigit($vColumn) Then 
        $vColumn = __WDXGetColumnIndexByNameInternal($aColumns, $vColumn)
        If @error Then Return SetError(@error, 0, "")
    Else
        If Not (($vColumn >= 0) And $vColumn < UBound($aColumns)) Then Return SetError($ERROR_ERRCOLUMN, 0, -1)
    EndIf

    Return $aColumns[$vColumn][$nInfoIndex]
EndFunc

; Return: zero-based column or unit index, 
; Error: -1 and sets @error
Func _WDXGetIndexByName(ByRef $aPlugin, $vColumn, $vUnit = Default)
    If Not _WDXIsValidHandle($aPlugin) Then Return SetError($ERROR_ERRHANDLE, 0, "")
    
    Local $aColumns = $aPlugin[$PLUGIN_COLUMNS]

    ; if field is given by name
    If Not StringIsDigit($vColumn) Then 
        $vColumn = __WDXGetColumnIndexByNameInternal($aColumns, $vColumn)
        If @error Then Return SetError(@error, 0, "")
    Else
        If Not (($vColumn >= 0) And $vColumn < UBound($aColumns)) Then Return SetError($ERROR_ERRCOLUMN, 0, -1)
    EndIf

    If $vUnit = Default Then Return $vColumn

    Local $aUnits = $aColumns[$vColumn][$COLUMN_UNIT]

    ; if unit is given by name
    If Not StringIsDigit($vUnit) Then 
        $vUnit = __WDXGetUnitIndexByNameInternal($aUnits, $vUnit)
        Return SetError(@error, 0, $vUnit)
    Else
        If Not (($vUnit >= 0) And $vUnit < UBound($aUnits)) Then Return SetError($ERROR_ERRUNIT, 0, -1)
    EndIf

    Return $vUnit
EndFunc

; Return: column value
; Error: empty string and sets @error
Func _WDXGetValue(ByRef $aPlugin, $sFileName, $vColumn, $vUnit = 0)
    If Not _WDXIsValidHandle($aPlugin) Then Return SetError($ERROR_ERRHANDLE, 0, "")

    Local $aColumns = $aPlugin[$PLUGIN_COLUMNS]

    ; if field is given by name
    If Not StringIsDigit($vColumn) Then 
        $vColumn = __WDXGetColumnIndexByNameInternal($aColumns, $vColumn)
        If @error Then Return SetError(@error, 0, "")
    Else
        If Not (($vColumn >= 0) And $vColumn < UBound($aColumns)) Then Return SetError($ERROR_ERRCOLUMN, 0, "")
    EndIf

    Local $aUnits = $aColumns[$vColumn][$COLUMN_UNIT]

    ; if unit is given by name
    If Not StringIsDigit($vUnit) Then 
        $vUnit = __WDXGetUnitIndexByNameInternal($aUnits, $vUnit)
        If @error Then Return SetError(@error, 0, "")
    Else
        If $vUnit = 0 Then
            ; allowed, do nothing
        ElseIf Not (($vUnit > 0) And $vUnit < UBound($aUnits)) Then 
            Return SetError($ERROR_ERRUNIT, 0, "")
        Endif
    EndIf

    Local $vFieldValue = __WDXGetFieldValue($aPlugin[$PLUGIN_HANDLE], $sFileName, $vColumn, $vUnit)
    Return SetError(@error, @extended, $vFieldValue)
EndFunc

; Return: array of all values
; Error: empty string and sets @error
Func _WDXGetValues(ByRef $aPlugin, $sFileName)
    If Not _WDXIsValidHandle($aPlugin) Then Return SetError($ERROR_ERRHANDLE, 0, "")

    Local $aColumns = $aPlugin[$PLUGIN_COLUMNS]

    Local $aFieldValues[UBound($aColumns)]
    For $i = 0 To UBound($aColumns) - 1
        $vFieldValue = __WDXGetFieldValue($aPlugin[$PLUGIN_HANDLE], $sFileName, $i, 0)
        Switch @error
            Case 0
                $aFieldValues[$i] = $vFieldValue
                ContinueLoop
            Case $ERROR_ERRVALUE
                $aFieldValues[$i] = _WDXValueAsString(@extended)
                ContinueLoop
            Case Else
                ExitLoop
        EndSwitch
    Next

    Return $aFieldValues
EndFunc

Func _WDXIsValidHandle(ByRef $aPlugin)
    Return IsArray($aPlugin) And ($aPlugin[$PLUGIN_HANDLE] <> 0) And IsArray($aPlugin[$PLUGIN_COLUMNS])
EndFunc

; Name by value
Func _WDXValueAsString($nValue)
    Switch $nValue
        Case $ft_nomorefields
            Return "ft_nomorefields" 
        Case $ft_numeric_32
            Return "ft_numeric_32" 
        Case $ft_numeric_64
            Return "ft_numeric_64" 
        Case $ft_numeric_floating
            Return "ft_numeric_floating" 
        Case $ft_date
            Return "ft_date" 
        Case $ft_time
            Return "ft_time" 
        Case $ft_boolean
            Return "ft_boolean" 
        Case $ft_multiplechoice
            Return "ft_multiplechoice" 
        Case $ft_stringA
            Return "ft_string"
        Case $ft_fulltext
            Return "ft_fulltext" 
        Case $ft_datetime
            Return "ft_datetime" 
        Case $ft_stringW
            Return "ft_stringw" 
        Case $ft_fulltextW
            Return "ft_fulltextw"
        Case $ft_comparecontent
            Return "ft_comparecontent"
        Case $ft_nosuchfield = -1
            Return "ft_nosuchfield"
        Case $ft_fileerror = -2
            Return "ft_fileerror"
        Case $ft_fieldempty = -3
            Return "ft_fieldempty"
        Case $ft_ondemand = -4        
            Return "ft_ondemand"
    EndSwitch    
EndFunc

#cs --------------------------------------------------------------------------------------------------------------------------
Internal usage only
#ce --------------------------------------------------------------------------------------------------------------------------

Func __WDXGetColumnIndexByNameInternal(ByRef $aColumns, $sColumnName)
    For $i = 0 To UBound($aColumns) - 1 
        If $sColumnName = $aColumns[$i][$COLUMN_NAME] Then Return $i
    Next
    Return SetError($ERROR_ERRCOLUMN, 0, -1)
EndFunc

Func __WDXGetUnitIndexByNameInternal(ByRef $aUnits, $sUnitName)
    For $j = 0 To UBound($aUnits) - 1
        If $aUnits[$j] = $sUnitName Then Return $j
    Next
    Return SetError($ERROR_ERRUNIT, 0, -1)
EndFunc

Func __WDXGetColumnsInternal($hPlugin)
    Local $aRet
    Local $aColumns[$MAX_COLUMNS][$COLUMN_MAXINDEX], $nColumnCount = 0
    For $i = 0 To $MAX_COLUMNS - 1 
        $aRet = DllCall($hPlugin, "int", "ContentGetSupportedField", "int", $i, "str", "", "str", "", "int", $MAX_BUFFER_SIZE) 
        If @error Then Return SetError($ERROR_ERRCALL, 0, 0)
        ; checking also for negative values due some wrong plugins
        If $aRet[0] = $ft_nomorefields Or $aRet[0] < 0 Then ExitLoop
        $aColumns[$i][$COLUMN_TYPE] = $aRet[0]
        $aColumns[$i][$COLUMN_NAME] = $aRet[2]
        If $aRet[3] = "" Then
            $aColumns[$i][$COLUMN_UNIT] = $aRet[3]
        Else
            $aColumns[$i][$COLUMN_UNIT] = StringSplit($aRet[3], "|", 2)
        EndIf
        $nColumnCount += 1
    Next
    If $nColumnCount = 0 Then Return SetError($ERROR_ERRZEROCOLUMNS, 0, 0)
    ReDim $aColumns[$nColumnCount][$COLUMN_MAXINDEX]
    Return $aColumns
EndFunc

Func __WDXGetFieldValue($hPlugin, $sFileName, $nFieldIdx, $nUnitIdx)
    Local $pBuffer = DllStructCreate("byte[" & $MAX_BUFFER_SIZE & "]")
    Local $aRet = DllCall($hPlugin, "int", "ContentGetValueW", "wstr", $sFileName, "int", $nFieldIdx, "int", $nUnitIdx, "ptr", DllStructGetPtr($pBuffer), "int", $MAX_BUFFER_SIZE, "int", 0)
    ; in case of error "function not found", try to load ANSI version
    If @error = 3 Then $aRet = DllCall($hPlugin, "int", "ContentGetValue", "str", $sFileName, "int", $nFieldIdx, "int", $nUnitIdx, "ptr", DllStructGetPtr($pBuffer), "int", $MAX_BUFFER_SIZE, "int", 0) 
    If @error Then Return SetError($ERROR_ERRCALL, 0, "")
    Switch $aRet[0]
        Case $ft_numeric_32 
            $pValue = DllStructCreate("int", DllStructGetPtr($pBuffer)) 
        Case $ft_numeric_64 
            $pValue = DllStructCreate("int64", DllStructGetPtr($pBuffer)) 
        Case $ft_numeric_floating 
            $pValue = DllStructCreate("double", DllStructGetPtr($pBuffer)) 
        Case $ft_stringa, $ft_multiplechoice
            $pValue = DllStructCreate("char[" & $MAX_BUFFER_SIZE & "]", DllStructGetPtr($pBuffer)) 
        Case $ft_stringw 
            $pValue = DllStructCreate("wchar[" & Int($MAX_BUFFER_SIZE / 2) & "]", DllStructGetPtr($pBuffer))
        Case $ft_fulltextw, $ft_fulltext
            $pValue = DllStructCreate(($aRet[0] = $ft_fulltextw ? "w" : "") & "char[" & $MAX_BUFFER_SIZE & "]", DllStructGetPtr($pBuffer))
            Local $sText = DllStructGetData($pValue, 1)
            If ($aRet[1] = $ft_fulltextw) Or ($aRet[1] = $ft_fulltext) Then $sText &= __WDXGetFieldValue($hPlugin, $sFileName, $nFieldIdx, $nUnitIdx)
            Return $sText
        Case $ft_date
            $pDate = DllStructCreate("word;word;word", DllStructGetPtr($pBuffer))
            Return StringFormat("%04d-%02d-%02d", _ 
                                DllStructGetData($pDate, 1), _ 
                                DllStructGetData($pDate, 2), _ 
                                DllStructGetData($pDate, 3))
        Case $ft_time
            $pTime = DllStructCreate("word;word;word", DllStructGetPtr($pBuffer))
            Return StringFormat("%02d:%02d:%02d", _ 
                                DllStructGetData($pTime, 1), _ 
                                DllStructGetData($pTime, 2), _ 
                                DllStructGetData($pTime, 3))
        Case $ft_datetime
            $pFileTime = DllStructCreate("dword;dword", DllStructGetPtr($pBuffer)) 
            $pLocalTime = DllStructCreate("dword;dword") 
            DllCall("Kernel32.dll", "int", "FileTimeToLocalFileTime", "ptr", DllStructGetPtr($pFileTime), "ptr", DllStructGetPtr($pLocalTime)) 
            $pValue = DllStructCreate("short;short;short;short;short;short;short;short") 
            DllCall("Kernel32.dll", "int", "FileTimeToSystemTime", "ptr", DllStructGetPtr($pLocalTime), "ptr", DllStructGetPtr($pValue)) 
            Return StringFormat("%04d-%02d-%02d %02d:%02d:%02d", _ 
                                DllStructGetData($pValue, 1), _ 
                                DllStructGetData($pValue, 2), _ 
                                DllStructGetData($pValue, 4), _ 
                                DllStructGetData($pValue, 5), _ 
                                DllStructGetData($pValue, 6), _ 
                                DllStructGetData($pValue, 7))
        Case $ft_boolean
            $pValue = DllStructCreate("boolean", DllStructGetPtr($pBuffer))
;        Case $ft_nosuchfield, $ft_fieldempty, $ft_ondemand ; non-fatal errors
;            Return SetError($ERROR_ERRVALUE, $aRet[0], "")
        Case Else
            Return SetError($ERROR_ERRVALUE, $aRet[0], "")
    EndSwitch 
    Return SetExtended($aRet[0], DllStructGetData($pValue, 1))
EndFunc

