#include <TCWDXPlugin.au3>

#pragma compile(FileDescription, Console WDX (Unicode))
#pragma compile(FileVersion, 1.3.0)
#pragma compile(LegalCopyright, 2016-2024 Dmitry Yudin)
#pragma compile(Console, true)
#pragma compile(Icon, uwdx.ico)

; Icon from "Flat Icon" set, flat-icons.com

#NoTrayIcon

If $CmdLine[0] = 0 Then _FatalError("No any parameters passed." & @CRLF & @CRLF & "Usage: " & @ScriptName & " [/C:codepage] [/I[:<ini path>]] <plugin.wdx> [testfile [column [unit]]]" & @CRLF)

; Handle switches
Global $iCodepage = 65001, $sTCIniPath
Global $aParams[1] = [0]
For $i = 1 To $CmdLine[0]
    Switch StringLeft($CmdLine[$i], 2)
        Case "/C"
            If StringMid($CmdLine[$i], 3, 1) = ":" Then
                $sCP = StringTrimLeft($CmdLine[$i], 3)
                If StringIsDigit($sCP) Then
                    $iCodepage = Number($sCP)
                    ContinueLoop
                EndIf
            EndIf
            _FatalError("Wrong codepage option.")

        Case "/I"
            If StringLen($CmdLine[$i]) = 2 Then
                $sTCIniPath = EnvGet("COMMANDER_INI_PATH") ; from TC 11.03
                ; but what if ini in profile?
                If Not $sTCIniPath Then $sTCIniPath = EnvGet("COMMANDER_PATH")
                If Not $sTCIniPath Then _FatalError("TC configuration not found. Please run uwdx from TC or provide wincmd.ini path.")
                $sTCIniPath &= "\wincmd.ini"
            ElseIf StringMid($CmdLine[$i], 3, 1) = ":" Then
                $sTCIniPath = StringTrimLeft($CmdLine[$i], 3)
            EndIf

            If FileExists($sTCIniPath) Then ContinueLoop
            _FatalError("Passed TC configuration path does not exists.")

    EndSwitch

    $aParams[0] += 1
    ReDim $aParams[$aParams[0] + 1]
    $aParams[$aParams[0]] = $CmdLine[$i]
Next

If $aParams[0] = 0 Then _FatalError("No required parameters passed.")

Global $aTypes[12] = ["Unknown", "Int32", "Int64", "Float", "Date", "Time", "Boolean", "MultipleChoice", "String", "FullText", "DateTime", "UnicodeString"]
Global $sFileName = ""
Global $nColumnIdx = 0
Global $nUnitIdx = 0
Global $sFormatStr

If $sTCIniPath Then
    Global $sPluginName = _GetIniPluginPath($sTCIniPath, $aParams[1])
    If @error = 1 Then
        _FatalError("Content plugins section not found.")
    ElseIf @error = 2 Then
        _FatalError("Plugin with name """ & $aParams[1] & """ is not found.")
    EndIf
Else
    Global $sPluginName = $aParams[1]
EndIf

If Not FileExists($sPluginName) Then _FatalError("Plugin file """ & $sPluginName & """ does not exists.")

Global $aPlugin = _WDXPluginOpen($sPluginName)
If @error Then _FatalError("Error opening plugin file, probably not WDX plugin or plugin architecture incompatible with this uwdx version.")

Opt("OnExitFunc", "ExitFunc")
Global $g_hKernel32 = DllOpen("kernel32.dll")
Global $g_hOutHandle = _GetStdHandle(-11)

DllCall("kernel32.dll", "bool", "SetConsoleOutputCP", "uint", $iCodepage)

If $aParams[0] = 1 Then
    $aColumns = _WDXGetColumns($aPlugin)
    ConsoleWrite("All columns: " & @CRLF & @CRLF)
    $nColPadSize = StringLen(UBound($aColumns))
    For $i = 0 To UBound($aColumns) - 1
        $sFormatStr = StringFormat("[%" & $nColPadSize & "s] -> %s (%s)\r\n", $i, $aColumns[$i][$COLUMN_NAME], _GetTypeString($aColumns[$i][$COLUMN_TYPE]))
        ConsoleWrite($sFormatStr)
        If IsArray($aColumns[$i][$COLUMN_UNIT]) Then
            $aUnits = $aColumns[$i][$COLUMN_UNIT]
            $nUnitPadSize = StringLen(UBound($aUnits))
            For $j = 0 To UBound($aUnits) - 1
                $sFormatStr = StringFormat("%" & $nColPadSize + 2 & "s [%" & $nUnitPadSize & "s] -> %s\r\n", "", $j, $aUnits[$j])
                ConsoleWrite($sFormatStr)
            Next
        EndIf
    Next
    Exit
EndIf

If $aParams[0] >= 2 Then
    ; some plugins fail if no full path to file passed (TC plugin API always passing full path)
    $sFileName = _GetFullPath($aParams[2])
    If Not FileExists($sFileName) Then _FatalError("Test file does not exists.")
EndIf

If $aParams[0] = 2 Then
    $aColumns = _WDXGetColumns($aPlugin)
    ConsoleWrite("All column values: " & @CRLF & @CRLF)
    For $i = 0 To UBound($aColumns) - 1
        $aUnits = $aColumns[$i][$COLUMN_UNIT]
        If IsArray($aUnits) Then
            ConsoleWrite($aColumns[$i][$COLUMN_NAME] & @CRLF)
            For $j = 0 To UBound($aUnits) - 1
                ConsoleWrite("    ")
                $sValue = _WDXGetValue($aPlugin, $sFileName, $i, $j)
                ConsoleWrite(_TextAlign($aUnits[$j], 28) & ": ")
                _WriteOutput($g_hOutHandle, $sValue, $iCodepage)
                ConsoleWrite(@CRLF)
            Next
        Else
            $sValue = _WDXGetValue($aPlugin, $sFileName, $i)
            ConsoleWrite(_TextAlign($aColumns[$i][$COLUMN_NAME], 32) & ": ")
            _WriteOutput($g_hOutHandle, $sValue, $iCodepage)
            ConsoleWrite(@CRLF)
        EndIf
    Next
    Exit
EndIf

If $aParams[0] >= 3 Then $nColumnIdx = $aParams[3]
If $aParams[0] = 4 Then $nUnitIdx = $aParams[4]

$sValue = _WDXGetValue($aPlugin, $sFileName, $nColumnIdx, $nUnitIdx)
If @error Then
    Switch @error
        Case $ERROR_ERRCOLUMN
            ConsoleWrite("Error in column name or index.")
        Case $ERROR_ERRUNIT
            ConsoleWrite("Error in unit name or index.")
        Case $ERROR_ERRZEROCOLUMNS
            ConsoleWrite("Error retrieving columns info, zero columns was returned.")
        Case $ERROR_ERRVALUE,
            ConsoleWrite("Plugin error when retrieving value.")
        Case $ERROR_ERRCALL
            ConsoleWrite("Error calling plugin function.")
        Case Else
            ConsoleWrite("Unknown error when retieving value.")
    EndSwitch
EndIf

_WriteOutput($g_hOutHandle, $sValue, $iCodepage)

Func _FatalError($sMessage)
    ConsoleWrite($sMessage)
    Exit
EndFunc

Func ExitFunc()
    _WDXPluginClose($aPlugin)
    DllClose($g_hKernel32)
EndFunc

Func _GetTypeString($nType)
    If $nType > UBound($aTypes) Then $nType = 0
    Return $aTypes[$nType]
EndFunc

Func _TextAlign($sString, $nMaxCount)
    Local $sSpaces = "          "
    Local $nLen = StringLen($sString)
    If $nLen < $nMaxCount Then
        For $i = 1 To Ceiling(($nMaxCount - $nLen) / 10)
            $sString &= $sSpaces
        Next
    EndIf
    Return StringLeft($sString, $nMaxCount)
EndFunc

Func _WriteOutput($hHandle, $sText, $iCodepage = 65001)
    Local $aResult = DllCall($g_hKernel32, "int", "WideCharToMultiByte", _
            "uint", $iCodepage, _
            "dword", 0, _
            "wstr", $sText, _
            "int", -1, _
            "ptr", 0, _
            "int", 0, _
            "ptr", 0, _
            "ptr", 0)
    If @error Or Not $aResult[0] Then Return SetError(@error + 20, @extended, "")

    Local $nMultiByteSize = $aResult[0]
    Local $tMultiByte = DllStructCreate("byte[" & $nMultiByteSize & "]")

    $aResult = DllCall($g_hKernel32, "int", "WideCharToMultiByte", _
            "uint", $iCodepage, _
            "dword", 0, _
            "wstr", $sText, _
            "int", -1, _
            "struct*", $tMultiByte, _
            "int", $nMultiByteSize, _
            "ptr", 0, _
            "ptr", 0)
    If @error Or Not $aResult[0] Then Return SetError(@error + 10, @extended, "")

    Local $pNumberOfCharsWritten = 0
    ; $nMultiByteSize - 1, strip termination zero
    DllCall($g_hKernel32, "bool", "WriteFile", _
            "handle", $hHandle, _
            "ptr", DllStructGetPtr($tMultiByte), _
            "dword", $nMultiByteSize - 1, _
            "dword*", $pNumberOfCharsWritten, _
            "ptr", 0)
EndFunc

Func _GetStdHandle($nIndex)
    Local $aRes = DllCall($g_hKernel32, "handle", "GetStdHandle", "dword", $nIndex)
    Return $aRes[0]
EndFunc

Func _GetFullPath($sFileName)
    Local $tBuffer = DllStructCreate("wchar[4096]")
    ; Important, buffer size in TCHARs, or function fail
    Local $aRet = DllCall("kernel32.dll", "dword", "GetFullPathNameW", _
                                          "wstr", $sFileName, _
                                          "dword", DllStructGetSize($tBuffer) / 2, _
                                          "ptr", DllStructGetPtr($tBuffer), _
                                          "ptr", 0)
    Return DllStructGetData($tBuffer, 1)
EndFunc


Func _GetIniPluginPath($sIniName, $sPluginName)
    ; extract plugin name (filename without ext)
    $sPluginName = StringRegExpReplace($sPluginName, "(?m)^.*?\\?([^\\]+)\.[^\.]+$", "\1")

    Local $aSect = IniReadSection(_TCHIniGetRedirectName($sIniName, "ContentPlugins"), "ContentPlugins")
    If @error Then Return SetError(1)

    For $i = 1 To $aSect[0][0]
        If StringIsDigit($aSect[$i][0]) And StringRegExpReplace($aSect[$i][1], "(?m)^.+?\\([^\\]+)\.[^\.]+$", "\1") = $sPluginName Then
            Return _TCHExpandEnv($aSect[$i][1])
        EndIf
    Next

    Return SetError(2)
EndFunc

; Return redirected ini
Func _TCHIniGetRedirectName($sIniName, $sSection)
    Local $sBasePath = "", $sIniError = "<error>"
    Local $sRedirect = IniRead($sIniName, $sSection, "RedirectSection", $sIniError)
    If $sRedirect <> $sIniError Then
        If StringInStr($sRedirect, "%") Then $sRedirect = _TCHExpandEnv($sRedirect)
        ; if path is local or UNC then return redirected path itself
        If StringInStr($sRedirect, ":") Or (StringLeft($sRedirect, 2) = "\\")  Then Return $sRedirect
        ; else assume redirect path is relative and extract source ini base path
        $sBasePath = StringRegExpReplace($sRedirect, "(.+\\)([^\\]+)", "\1")
        $sIniName = $sBasePath & $sRedirect
    EndIf
    Return $sIniName
EndFunc

Func _TCHExpandEnv($sText)
    Return DllCall("Kernel32.dll", "bool", "ExpandEnvironmentStringsW", "wstr", $sText, "wstr", "", "uint", 32768)[2]
EndFunc
